package other

import scala.util.parsing.combinator._
import sun.java2d.pipe.SpanShapeRenderer.Simple
import other.Tudo.rotulos
import com.sun.org.apache.xml.internal.serializer.ToStream.BoolStack

case class ids_t(ids:List[String], tipo: String)
case class dec_param(byRef:Boolean, ids:List[String], tipo:String)
    
class Arith extends JavaTokenParsers {
  //CONTROLE etc.
  val fazTudo = Tudo
  var nivelLexico = 0
  val rotStack = Tudo.proc.RotStack
  
  val stackbooltipo = collection.mutable.Stack[String]()
  val stacktipo = collection.mutable.Stack[String]()
  val procStack = collection.mutable.Stack[Entrada]()
  val paramCounterStack = collection.mutable.Stack[Int]()
  
  var eh_atrib = false //esta dentro de uma atribuicao

  def eh_param = procStack match {
    case stack if ! stack.isEmpty => true
//      stack.top.isInstanceOf[Procedimento] || stack.top.isInstanceOf[Funcao]
    case _ => false
  }
  
  //PARSER
  object PascalParser {

    implicit class isolavel(s:String) {
      def iso = (s + "\\s").r ^^ { _.take(s.size) }
    }
    
    //PROGRAM
    def program  = {
      def programInOut = "(" ~> repsep(ident,",") <~ ")"
      fazTudo.iniciaPrograma
      "program".iso ~> ident ~ opt(programInOut) <~ ";" ~ bloco ~ "." ^^
        { _ => fazTudo.encerraPrograma }
    }
    
    def bloco:Parser[Any] =
        (opt(dec_rotulos) ~
            (opt(dec_tipos) ^^ {
              case Some(tipos) =>
                //Add to table
                TabelaSimb.insere(tipos.map{ t => Tipo(t) })
              case None => //fine
            })~
            opt(dec_vars) ^^ 
            { _ => 
              if (nivelLexico==0) Tudo.aplica("DSVS R00");
              else Tudo.aplica("DSVS "+Tudo.proc.RotStack.top)
              nivelLexico+=1 
        }) ~
        (opt(dec_subrotinas) ^^ 
          { _ => 
            nivelLexico-=1
            if (nivelLexico==0) Tudo.aplica("R00: NADA") 
            else
              if (Tudo.MEPA.last == "DSVS "+Tudo.proc.RotStack.top) {
                //DSVS R003; R003: seguidos!
                Tudo.rotulos.--
                Tudo.proc.RotStack.pop()
                Tudo.MEPA.remove(Tudo.MEPA.size - 1)
              } else
                Tudo.aplica(Tudo.proc.RotStack.pop()+": NADA")
          }
        ) ~ comando_composto

    //ROTULOS
    def dec_rotulos = "label".iso ~> """\d+""".r ~ rep("," ~> """\d+""".r) <~ ";" ^^
      { case x~xs => fazTudo.insereRotulos(x::xs, nivelLexico) }
   
    //TIPOS
    def dec_tipos = "type".iso ~> (def_tipo <~ ";") ~  rep(def_tipo <~ ";" ) ^^ { mkList }
    
    def def_tipo = ident <~ "=" ~ "integer"
    
    def tipo = ident
    
    //VARIAVEIS
    def dec_vars = "var".iso ~> dec_var ~ rep(";" ~> dec_var) <~ ";" ^^
      { case xss => fazTudo.insereIntVars(mkList(xss), nivelLexico) } 
    def dec_var  = (lista_ident <~ ":") ~ tipo ^^ { case ids ~ t => ids_t(ids, t)}
    def lista_ident = ident ~ rep( "," ~> ident ) ^^ { mkList }
    
    //SUB-ROTINAS
    def dec_subrotinas = rep( (dec_proc | dec_func) <~ bloco ~ ";" ^^ 
        {  e => fazTudo.proc.saiProcFunc(nivelLexico); e} )
    
    def dec_proc = "procedure".iso ~> ident ~ opt(params_formais) <~ ";" ^^
      { case id ~ lstDecParam => fazTudo.proc.entraProc(id, lstDecParam, nivelLexico) }
    def dec_func = "function".iso  ~> ident ~ opt(params_formais) ~ (":" ~> tipo) <~ ";" ^^
      { case id ~ lstDecParam ~ tipo => fazTudo.proc.entraFunc(id, lstDecParam, nivelLexico, tipo) }
    
    def params_formais = "(" ~> params ~ rep(";" ~> params) <~ ")" ^^ { mkList }
    def params = opt("var".iso) ~ lista_ident ~ (":" ~> tipo) ^^ { case v ~ ids ~ tipo => dec_param(v.isDefined, ids, tipo)}
    
    //COMANDOS
    def comando_composto:Parser[Any] = "begin".iso ~ comando ~ rep(";" ~ comando) ~ "end"
    def comando = opt("""\d+""".r <~ ":" ^^ { rot => fazTudo.entraRotulo(rot,nivelLexico) }) <~ comando_sem_rotulo
    def comando_sem_rotulo:Parser[Any] = desvio | atribuicao | repeticao |
        condicional | comando_composto | chamaId | expr_simples
        
    def atribuicao = (ident ^^ {e => eh_atrib = true; e}) <~ ":=" ~ expr_simples ^^
        { 
          case id => 
            eh_atrib = false
            verificaTipoAtrib(id)
            fazTudo.atribuicao(id) }
    
    
    def verificaTipoAtrib (idAtrib:String) {
      val e = TabelaSimb.getById(idAtrib).get
      val boolDistinct = stackbooltipo.distinct
      if (boolDistinct.size!=1 && !(boolDistinct.size==2 && boolDistinct.contains("integer")) )
      {
        println("Erro de tipagem na expressão da atribuição")
        sys.exit(1)        
      }
      
      val tipoExpr = boolDistinct.size match {
        case 1 => boolDistinct.pop()
        case 2 => boolDistinct.top match {
                    case "integer" => boolDistinct.pop(); boolDistinct.pop()
                    case _ => boolDistinct.pop()
                  }
      }
      boolDistinct.clear()
      
      val tipoAtrib = e match {
        case v:Var    => v.tipo  
        case p:Param  => p.tipo
        case f:Funcao => f.tipo
        case _ => println("Encontrei um procedimento em uma atribuição?"); "integer"
      }
      //assert tipos
      if (tipoExpr!="integer" && tipoAtrib!="integer" && tipoExpr!=tipoAtrib) {
          println("Erro de tipo: Atribuição recebendo tipo diferente. (["+tipoAtrib+"] := ["+tipoExpr+"])")
          sys.exit(1)
      } //Otherwise ok!
    }
    
    def desvio = "goto".iso ~> """\d+""".r ^^ { rot => fazTudo.gotoRotulo(rot,nivelLexico) }
    
    def condicional = 
      ("if".iso ~ boolexpr ^^ { e =>
        val r = Tudo.rotulos.geraRotulo
        Tudo.aplica("DSVF "+r)
        rotStack.push(r)
        e
      }) ~ 
      "then".iso ~
      (comando_sem_rotulo ^^ { _ => Tudo.aplica(rotStack.pop+": NADA") }) ~
      opt(
          ("else".iso ^^ { _ =>
            val gotoElse = Tudo.MEPA.remove(Tudo.MEPA.size - 1)//.takeWhile {_!=':'}
            //rotStack.push(fim_if)
            rotStack.push(rotulos.geraRotulo)
            Tudo.aplica("DSVS "+rotStack.top) //Vai pro fim do if
            Tudo.aplica(gotoElse)
          }) 
          ~ (comando_sem_rotulo ^^ { _ => Tudo.aplica(rotStack.pop+": NADA") })
      )
       
    def repeticao = 
      (("while" ^^ { _=> 
          val rotLoop = rotulos.geraRotulo
          Tudo.aplica(rotLoop+": NADA")
          rotStack.push(rotLoop)
      }) ~ boolexpr ~ "do" ^^ { _ => 
          val rotFim = rotulos.geraRotulo
          Tudo.aplica("DSVF "+rotFim)
          rotStack.push(rotFim)
      }) ~ comando_sem_rotulo ^^ { _=>
          val rotFim = rotStack.pop
          val rotLoop = rotStack.pop
          Tudo.aplica("DSVS "+rotLoop)
          Tudo.aplica(rotFim+": NADA")
      }
    
    //EXPRESSOES BOOLEANAS
    def boolexpr:Parser[Unit] =
      (
        boolexpr_simples ~> rep ( ("and"|"or") <~ boolexpr )
      ) ^^
      { lst => lst.reverse.foreach { case "and" => fazTudo.aplica("CONJ")
                                     case "or"  => fazTudo.aplica("DISJ")}
      }
    
    def boolexpr_simples =
      (
        expr_simples ~> relacao <~ expr_simples | 
        "(" ~ expr_simples ~> relacao <~ expr_simples ~ ")"
      ) ^^
          { e =>
            if (stackbooltipo.size < 2)
              println("stackbooltipo size less than 2")
              
            val distinct = stackbooltipo.distinct
            if (distinct.size==1 || (distinct.size==2 && distinct.contains("integer")) )
            {
              //fine
              stackbooltipo.clear()
            }
            else {
                distinct.foreach{println}
                println("Erro de tipo: dois lados da expressão booleana com tipos diferentes")
                sys.exit(1)
            }
            e match {
              case "="  => fazTudo.aplica("CMIG")
              case "<>" => fazTudo.aplica("CMDG")
              case "<=" => fazTudo.aplica("CMEG")
              case ">=" => fazTudo.aplica("CMAG")
              case "<"  => fazTudo.aplica("CMME")
              case ">"  => fazTudo.aplica("CMMA")
              case _ => sys.exit(1)
            }
          }
    def relacao = "<="|">="|"="|"<>"|"<"|">"
    
    //EXPRESSOES
    
    def lista_expr = 
      (expr_simples ^^ { foreachExpr }) ~
        rep("," ~> (expr_simples ^^ { foreachExpr })) ^^ { mkList }
    
    def foreachExpr(o:Object) = {
      procStack.top.id  match {
        case "write" => fazTudo.aplica("IMPR")
        case "read"  => //ok
        case _ =>
      }
      val isByRef = Tudo.isAtArgByRef()
      inc_pCount
      // Check for byRef args
      o match {
        case s:String => //Okey
        case o => 
          if ( procStack.top.id=="read" )
            assert(false, "Função read recebendo uma expressão!")
          else if (isByRef) 
            assert(false, "Parametro por referencia recebendo valor! Funcão/procedimento: " + procStack.top.id)
          //else println("Okey")
      }
      o
    }
    
    def inc_pCount {
      paramCounterStack.push(1+paramCounterStack.pop)
    }
    
    def expr_simples:Parser[Object] = termo ~ rep( ("+"|"-") ~ termo ) ^^
    { e =>
      stackbooltipo foreach println;
      //TIPO
      val distinct = stacktipo.distinct
      if ( distinct.size == 1 || ( distinct.size == 2 && distinct.contains("integer") ) )
      {
        if (distinct.size==2)
          //Less priority for integer
          if (stacktipo.top != "integer") stackbooltipo.push( stacktipo.pop() )
          else { stacktipo.pop(); stackbooltipo.push( stacktipo.pop() ) }
        else
          stackbooltipo.push( stacktipo.top )
      } else {
        println("Erro de tipo: Expresão com tipos diferentes")
        sys.exit(1)
      }
      stacktipo.clear()
      //END
      e match {
        case t ~ lst if ! lst.isEmpty =>
          lst.reverse.foreach {
            case ("+" ~ t) => fazTudo.aplica("SOMA")
            case ("-" ~ t) => fazTudo.aplica("SUBT")
            case _ => println("op???")
          }
          new AnyRef
        case t ~ Nil => t 
      }
    }
    
    def termo = fator ~ rep( ("*"|"div"|"/") ~ fator ) ^^
      { e =>
        e match {
          case fat~lst if ! lst.isEmpty =>
            lst.map{ case ~(str,_) => str }.reverse.foreach {
              case "*" => fazTudo.aplica("MULT")
              case "div"|"/" => fazTudo.aplica("DIVI")
            }
            new AnyRef
          case fat ~ Nil => fat
        }
      }
    def fator = ( "("~expr_simples~")" | chamaId | numero ) ^^ { case p:String => p; case _ => new AnyRef}
    
    def getParamsCount(e:Entrada) = e match {
      case f:Funcao => f.paramInfo.size
      case p:Procedimento => p.paramInfo.size
    }
    
  
  def ehProcFun(id:String) = {
    TabelaSimb.getById(id) match {
      case Some(e) => 
        e.isInstanceOf[Procedimento] || e.isInstanceOf[Funcao]
      case None => false
    }
  }
  
  def assert(bool: => Boolean, errMsg:String) {
    if ( ! bool ) {
      println("Erro: " + errMsg)
      sys.exit(1)
    }
  }
    
  def entraId(id:String) 
  {
    TabelaSimb.getById(id) match {
      case Some(p) if p.isInstanceOf[Param] =>  stacktipo.push(p.asInstanceOf[Param].tipo)
      case Some(v) if v.isInstanceOf[Var]=>  stacktipo.push(v.asInstanceOf[Var].tipo)
      case Some(f) if f.isInstanceOf[Funcao]=>  stacktipo.push(f.asInstanceOf[Funcao].tipo)
      case _ => 
    }
    if (ehProcFun(id)) {
      if ( TabelaSimb.getById(id).get.isInstanceOf[Funcao] )
        if (eh_atrib) Tudo.aplica("AMEM 1")
      procStack.push(TabelaSimb.getById(id).get)
      paramCounterStack.push(0)
      
    } else if (id=="write" || id=="read") {
      
      procStack.push(new Entrada(id))
      paramCounterStack.push(0)
      id match {
        case "write" =>
        case "read"  =>
      }
      
    }
  }
  
  def saiId(id:String)
  {
    
    if (id=="write" || id=="read") { 
      procStack.pop
      return
    }
    if (!procStack.isEmpty && procStack.top.id == "read"){
      fazTudo.aplica("LEIT")
      if (id != "read") {
        var tipo = ""
        TabelaSimb.getById(id) match  {
          case Some(s) => s match {
            case v:Var   => tipo = v.tipo
            case p:Param => tipo = p.tipo
            case _ =>
          }
          case _ => 
        }
        if (tipo != "") {
          if (tipo != stackbooltipo.pop()) {
            println("Erro de tipo: atribuição com tipos diferentes")
            sys.exit(1)
          }
        }
        
        fazTudo.atribuicao(id)
      }
      return
    }

      
    if ( ehProcFun(id) ) {
      val proc = procStack.pop
      val nParams = paramCounterStack.pop
      assert(nParams == getParamsCount(proc), 
          "Número de parametros incorretos na chamada da função/procedimento "+id)
    }

    fazTudo.geraMepaDoId(id, nivelLexico)
  }
  
  def chamaId =
   (ident ^^ { id => entraId(id); id }) <~ opt("(" ~ lista_expr ~ ")") ^^ 
   { id => saiId(id); id }
    
  def numero = wholeNumber ^^ { case x => stacktipo.push("integer");fazTudo.literal(x.toInt) }
    
  }
}

object MyParser extends Arith {
  
  def main(args: Array[String]) {
    if (args.size != 1) {
      println("./compilador [ArqEntrada]")
      sys.exit(1)
    }
    val pascalSource = io.Source.fromFile(args(0)).getLines().mkString("\n")
    try 
      println(parseAll(PascalParser.program, pascalSource))
    catch { case e : Exception => println("Erro no programa"); sys.exit(1) }
    val mp = Tudo.MEPA
    val pw = new java.io.PrintWriter("MEPA")
    mp foreach pw.println
    pw.close  
  }
  

//TODO: function return as parameter/atribution
}
