package other
import scala.annotation.tailrec

class Entrada(val id:String)
case class Funcao (override val id:String, paramInfo:List[Param], desloc:Int,
    nvLex:Int, rotEntrada:String, tipo:String = "integer") extends Entrada(id)
case class Procedimento(override val id:String, paramInfo:List[Param],
    nvLex:Int, rotEntrada:String) extends Entrada(id)
case class Param(override val id:String, desloc: Int, byRef:Boolean, nvLex:Int, tipo:String = "integer") extends Entrada(id)
case class Var(override val id:String, desloc: Int, nvLex:Int, tipo:String = "integer") extends Entrada(id)
case class Label(override val id:String, rotulo:String, nvLex:Int, var used:Boolean = false, var put:Boolean = false) extends Entrada(id)
case class Tipo(override val id:String) extends Entrada(id)

object TabelaSimb {
  
  private var table = List[Entrada]()
  
  def hasNext = table.size>0
  def top = table.head
  def drop = table = table.tail
  
  def getTable = table
  def setTable(t:List[Entrada]) = table = t
  
  def getById(id:String) = {
    table.find { e => e.id == id }
  }
  
  def getLastProc(nvLex:Int) = {
    table.find {
      case p:Procedimento => if (p.nvLex==nvLex) true else false
      case f:Funcao => if (f.nvLex==nvLex) true else false
      case _ => false
    }.get
  }
  
  @tailrec 
  def temParamOuVarComMesmoId(nome:String, rest:List[Entrada] = table):Boolean = { //No escopo da ultima funcao/proc
    val e = getById(nome)
    rest match {
      case (p:Procedimento) :: xs => false 
      case (f:Funcao) :: xs       => false
      case (v:Var) :: xs   => if (v.id == nome) true else temParamOuVarComMesmoId(nome, xs)
      case (p:Param) :: xs => if (p.id == nome) true else temParamOuVarComMesmoId(nome, xs)
      case _ => false
    }
  }
  
  def existeTipo(tipo:String):Boolean = {
    if (tipo == "integer") return true
    table.exists { e => e match {
      case t:Tipo =>
        if (t.id == tipo) true
        else false
      case _ => false
    }}
  }
  
  @tailrec
  def numVarsLocais(nvLex:Int, count:Int=0, rest:List[Entrada] = table):Int = rest match {
    case (v:Var) :: xs => numVarsLocais(nvLex, count+1, xs)
    case (p:Procedimento) :: xs => if (p.nvLex>nvLex) numVarsLocais(nvLex,count,xs) else count 
    case (f:Funcao) :: xs => if (f.nvLex>nvLex) numVarsLocais(nvLex,count,xs) else count 
    case _ => count
  }

  def insere(e:Entrada) {
    table = e :: table
  }
  
  def insere(lst:List[Entrada]) {
    table = lst.reverse ::: table
  }
  
  def print = table.reverse foreach println
}