package other

import scala.annotation.tailrec

object Tudo {
  import collection.mutable.ListBuffer
  val MEPA = ListBuffer[String]()  
  
  object rotulos {
    private var num = 0
    def -- = num-=1 
    def geraRotulo = {
      num += 1
      "R"+ (
          num match {
            case dig if num<10  => "0"+dig
            case dezena if num<100 => ""+dezena
            case maior if num>=100 => "" + maior
          }
       )
    }
  }
  
  def insereIntVars(vars_t: List[ids_t], nivelLexico: Int) = {
    
    //Garante que os tipos existem
    val lst_tipos = vars_t.map { case ids_t(_,t) => t }
    val tipo_invalido = lst_tipos.distinct.find { t => ! TabelaSimb.existeTipo(t) } //Ache um que não existe
    
    tipo_invalido match {
      case Some(invalido) => //Erro
        println("Erro: Tipo não declarado na declaração de variáveis: "+invalido)
        sys.exit(1)
      case None => //Ok
    }
    //end
    
    //Garante nenhuma repetição de id
    val listaIds = vars_t.map{ case ids_t(ids,_) => ids }.flatten
    val totalAMEM = listaIds.size
    
    if (listaIds.distinct.size != listaIds.size) {
        println("Erro: declaração de variáveis:")
        println("  identificadores já declarados para parametros ou var: "+
            listaIds.diff(listaIds.distinct).mkString(" "))
        sys.exit(1)
    }
    //end
    
    var desloc = 0
    vars_t.foreach { case ids_t(varsId,tipo) =>
        TabelaSimb insere varsId.map { id =>
            desloc += 1
            Var(id, desloc-1, nivelLexico,tipo)
        }
    }
    aplica("AMEM " +totalAMEM)
    
  }

  def geraMepaDoId(id: String, nvLexAtual:Int): Unit =
    try
      geraMepaDoId_code(id,nvLexAtual)
    catch {
      case e:IndexOutOfBoundsException =>
        val id = MyParser.procStack.pop.id
        println("Erro: Número de parametros"+
            " incorretos na chamada da função/procedimento "+id)
        sys.exit(1)
    }
    
  def isAtArgByRef() = {
    MyParser.procStack.top match {
      case p:Procedimento => p.paramInfo(MyParser.paramCounterStack.top).byRef
      case f:Funcao => f.paramInfo(MyParser.paramCounterStack.top).byRef
      case o => false //write and read will get here
    }
  }
  
  private def geraMepaDoId_code(id: String, nvLexAtual:Int): Unit = {
    TabelaSimb.getById(id) match {
      case Some(it) => it match {
        case v:Var => 
          if (MyParser.eh_param) {
            val isByRef = isAtArgByRef()
            if (isByRef)
              aplica("CREN "+v.nvLex+", "+v.desloc)              
            else
              aplica("CRVL "+v.nvLex+", "+v.desloc)
          } else
            aplica("CRVL "+v.nvLex+", "+v.desloc)
        case p:Param => 
          if (MyParser.eh_param) {
            val isByRef = isAtArgByRef()
            if (isByRef)
              if (p.byRef) aplica("CRVL "+p.nvLex+", "+p.desloc)
              else         aplica("CREN "+p.nvLex+", "+p.desloc)              
            else
              if (p.byRef) aplica("CRVI "+p.nvLex+", "+p.desloc) 
              else         aplica("CRVL "+p.nvLex+", "+p.desloc)
          } else
            if (p.byRef) aplica("CRVI "+p.nvLex+", "+p.desloc)
            else         aplica("CRVL "+p.nvLex+", "+p.desloc)
          
        case p:Procedimento => aplica("CHPR "+p.rotEntrada+", "+nvLexAtual)
          
        case f:Funcao => aplica("CHPR "+f.rotEntrada+", "+nvLexAtual)
          
        case other =>
          println("Erro(1): Identificador Inválido: "+ id)
          sys.exit(1)
      }
      case None =>
          println("Erro(2): Identificador Inválido: "+ id); 
          sys.exit(1)
    }
  }

  
  object proc {
   
    val RotStack = collection.mutable.Stack[String]()
   
    
    //DRY!
    def entraProcFun(nivelLexico: Int) = {
      val rotulo = rotulos.geraRotulo //rotuloEntrada
      val rotuloBegin = rotulos.geraRotulo
      RotStack.push(rotuloBegin) //Begin
      if (nivelLexico>1) { 
        val rotDesvio = TabelaSimb.getLastProc(nivelLexico-1) match {
          case p:Procedimento => p.rotEntrada
          case f:Funcao => f.rotEntrada
          case o => println("====> "+o)
        }
      }
      aplica(rotulo+": ENPR "+nivelLexico)
      rotulo
    }
    
    def entraProc(id:String, dec_ps: Option[List[dec_param]], nivelLexico:Int): Unit = {
      val rotulo = entraProcFun(nivelLexico)
      
      dec_ps match {
        case Some( lstParams ) if ! lstParams.isEmpty =>
          val ps = retornaParamsLst(lstParams, nivelLexico)
          TabelaSimb.insere( Procedimento(id, ps, nivelLexico, rotulo) )
          ps.foreach { param =>
            if (TabelaSimb.temParamOuVarComMesmoId(param.id)) {
              println("Erro: Parâmetros com mesmo nome no procedimento "+id+": "+ param.id)
              sys.exit(1)
            }
            TabelaSimb.insere(param) 
          }
        case None =>
          TabelaSimb.insere( Procedimento(id, List.empty, nivelLexico, rotulo) )
      }
    }
    
    def entraFunc(id:String, dec_ps: Option[List[dec_param]],nivelLexico:Int, tipo:String):Unit =
    {
      val rotulo = entraProcFun(nivelLexico)
      if ( ! TabelaSimb.existeTipo(tipo) ){
        
      }
     
       
      dec_ps match {
        case Some( lstParams ) if ! lstParams.isEmpty =>
          val ps = retornaParamsLst(lstParams, nivelLexico)
          val desloc = -4-ps.size
          TabelaSimb.insere( Funcao(id,ps,desloc,nivelLexico,rotulo,tipo) )
          ps.foreach { param =>
            if (TabelaSimb.temParamOuVarComMesmoId(param.id)) {
              println("Erro: Parâmetros com mesmo nome na função "+id+": "+ param.id)
              sys.exit(1)
            }
            TabelaSimb.insere(param)
          }
        case None =>
          TabelaSimb.insere( Funcao(id, List.empty, -4, nivelLexico, rotulo) )
      }
    }
    
    def retornaParamsLst(params:List[dec_param], nivelLexico:Int):List[Param] =  {
      var desloc = - 3 - params.map{ _.ids.size }.sum
      
      //Verifica tipos
      var todosTipos = params.map{ _.tipo }.distinct
      val invalido = todosTipos.find { t => ! TabelaSimb.existeTipo(t) }
      
      invalido match {
        case Some(inv) => //Erro
          println("Erro: Tipo não declarado na declaração de parametros: "+inv)
          sys.exit(1)
        case None => //Ok
      }
      //end
      
      params.flatMap {
        case dec_param(byRef, idlst, tipo) => 
          idlst map { id =>
            desloc +=1
            Param(id,desloc-1,byRef,nivelLexico,tipo)
          }
      }
    }
    
    @tailrec def contaVariaveis(nvLex:Int, count:Int = 0):Int = TabelaSimb.hasNext match {
      case true => TabelaSimb.top match {
        case v:Var => TabelaSimb.drop; contaVariaveis(nvLex,count+1)
        case p:Procedimento => if (p.nvLex>nvLex) {TabelaSimb.drop; contaVariaveis(nvLex,count)} else count
        case f:Funcao => if (f.nvLex>nvLex) {TabelaSimb.drop; contaVariaveis(nvLex,count)} else count
        case t:Tipo => TabelaSimb.drop; contaVariaveis(nvLex,count)
        case r:Label =>
          if (! r.used) {println("Warning: rótulo nunca usado: "+ r.id)}
          TabelaSimb.drop
          contaVariaveis(nvLex,count)
        case _ => count
      }
      case false => count
    }
    
    def saiProcFunc(nivelLexico:Int) {
      
      val numVariaveis = contaVariaveis(nivelLexico)
      if (numVariaveis>0)
        aplica("DMEM "+numVariaveis)
      
      val table = TabelaSimb.getTable
      TabelaSimb.setTable(
          table.dropWhile {
            case p:Procedimento => if (p.nvLex>nivelLexico) true else false
            case f:Funcao       => if (f.nvLex>nivelLexico) true else false
            case t:Tipo         => true
            case r:Label =>
              if (r.used) { //Se rótulo nunca usado
                println("Warning: rótulo nunca usado: "+ r.id)
                //sys.exit(1)
              }
              true
            case _ => true
          }
      )
      
      TabelaSimb.top match {
        case p:Procedimento => aplica("RTPR "+ p.nvLex +", "+p.paramInfo.size)
        case f:Funcao       => aplica("RTPR "+ f.nvLex+", "+f.paramInfo.size )
        case e => println(e); println("Nao eh instancia proc nem func-> ERRO"); sys.exit(2)
      }
    }
  }
  
  //ROTULOS
  def insereRotulos(lstRot: List[String], nivelLexico: Int) = {
    lstRot.foreach { 
      id => TabelaSimb insere ( Label(id,rotulos.geraRotulo,nivelLexico) )
    }
  }
  
  def entraRotulo(rotuloLogico:String,nvLex:Int) {
    val rot = TabelaSimb.getById(rotuloLogico)
    rot match {
      case Some(e) => e match {
        case r:Label => 
          if (r.put) {
            println("Erro: Rotulo "+r.id+" posto mais de uma vez!")
            sys.exit(1)
          } else {
            r.put = true
            //r.used = true goto should do that
            aplica(r.rotulo+": ENRT "+nvLex+", "+TabelaSimb.numVarsLocais(nvLex))
          }
        case _ => println("Erro: Rótulo "+rotuloLogico+" não declarado!"); sys.exit(1)
      } 
      case None => println("Erro: Rótulo "+rotuloLogico+" não declarado!"); sys.exit(1)
    }
  }

  def iniciaPrograma {
    aplica("INPP")
  }

  def encerraPrograma = {
    val countVars = proc.contaVariaveis(0)
    if (countVars>0) { aplica("DMEM "+countVars) }
    aplica("PARA")
  }

  def atribuicao(id: String) = {
    val e = TabelaSimb.getById(id)
    e match {
      case Some(e) => e match {
        case f:Funcao => aplica("ARMZ "+f.nvLex+", "+f.desloc)
        case v:Var => aplica("ARMZ "+v.nvLex+", "+v.desloc)
        case p:Param => 
          if (p.byRef) aplica("ARMI "+p.nvLex+", "+p.desloc) else aplica("ARMZ "+p.nvLex+", "+p.desloc)
        case other => println("Atribuição a "+id+" é inválida!"); sys.exit(1)
      }
      case None => println("identificador inválido: "+id); sys.exit(1)
    }
  }

  def literal(num: Int) = {
    aplica("CRCT "+num) 
  }

  def aplica(cmdMEPA: String) : Unit = 
    MEPA += (if ( ! cmdMEPA.contains(":") ) "     " else "") + cmdMEPA

  def paplica(cmdMEPA: String) = {
    println(cmdMEPA)
    aplica(cmdMEPA)
  }
  
  def gotoRotulo(rot: String, nivelLexico: Int) {
    val r = TabelaSimb.getById(rot)
    r match {
      case Some(s) => s match {
        case r:Label =>
          r.used = true
          aplica("DSVR "+r.rotulo+", "+r.nvLex+", "+nivelLexico)
        case _ => println("goto para rótulo não declarado! (goto "+rot+")"); sys.exit(1)
      }
      case None => println("goto para rótulo não declarado! (goto "+rot+")"); sys.exit(1)
    }
  }
  
//  def reservadas = List("program", "label", "var", "procedure", "function","begin","end",
//    "if", "then", "else", "while", "do", "or", "and","div","not")
  
}